/* client.cpp
 *
 * This is an ASCII client for the Borbaman game.
 */

/************
 * INCLUDES *
 ************/ 

// standard includes
#include <stdlib.h>
#include <ncurses.h>
#include <sys/time.h>

// C++ includes
#include <vector>
using namespace std;

// project includes
#include "game_constants.h"
#include "client.h"


/*************
 * CONSTANTS *
 *************/

// row offset in the terminal to start displaying the map -- this must be at 
// least 3
const unsigned int MAP_ROW_OFFSET = 5;

// column offset in the terminal to start displaying the map
const unsigned int MAP_COL_OFFSET = 10;

// colors for things on the map
const short CLR_EMPTY    = 1;
const short CLR_BRICK    = 2;
const short CLR_WALL     = 3;
const short CLR_PLAYER   = 4;
const short CLR_BOMB     = 5;
const short CLR_PLAYBOMB = 6;
const short CLR_FIRE     = 7;
const short CLR_PNUM     = 8;

// number of microseconds between updates with server
const long PERIOD        = 50000;  // 50 ms


/***********
 * GLOBALS *
 ***********/

// current game state received from the server
struct GameState gs;

// the player ID for this client
unsigned long player_id = 0;

// the server Number this client is connected to
int server_id = 0;


/**************
 * PROTOTYPES *
 **************/

static int get_player_at(unsigned int x, unsigned int y);
static bool initialize();
static void cleanup();
static void draw_map();
static void draw_map_border();
static void draw_players();
static void draw_bombs();
static void draw_scores();
static void draw_sequence_number(unsigned long num);
void set_next_time(struct timeval *next);
bool period_expired(const struct timeval *next);


/***********
 * HELPERS *
 ***********/

// returns the number of the player that is located at (x,y) or -1 if no one
// is at that location
static int get_player_at(unsigned int x, unsigned int y) {
  for (int i = 0; i < NUM_PLAYERS; i++) {
    if ((gs.players[i].x_coord == x) && (gs.players[i].y_coord == y)) {
      return i;
    }
  }
  return -1;
}

// Initializes the client and returns false if successful, otherwise true if an
// error occurred.
static bool initialize() {
  
  // puts the client into curses mode
  initscr();

  // enable colors
  if (start_color() == ERR)
    return true;

  // prevents characters from being echoed to the screen as the user types
  if (noecho() == ERR)
    return true;

  // don't buffer key input
  if (cbreak() == ERR)
    return true;

  // enable extended keys like the arrow keys and function keys
  if (keypad(stdscr, TRUE) == ERR)
    return true;

  // make the cursor invisible
  if (curs_set(0) == ERR)
    return true;

  // make getch() non-blocking
  if (nodelay(stdscr, true) == ERR) 
    return true;

  // create the game colors
  init_pair(CLR_EMPTY,    COLOR_BLACK,  COLOR_BLACK);
  init_pair(CLR_BRICK,    COLOR_YELLOW, COLOR_BLACK);
  init_pair(CLR_WALL,     COLOR_BLACK,  COLOR_WHITE);
  init_pair(CLR_PLAYER,   COLOR_CYAN,   COLOR_BLACK);
  init_pair(CLR_BOMB,     COLOR_WHITE,  COLOR_RED);
  init_pair(CLR_PLAYBOMB, COLOR_CYAN,   COLOR_RED);
  init_pair(CLR_FIRE,     COLOR_YELLOW, COLOR_RED);
  init_pair(CLR_PNUM,     COLOR_RED,    COLOR_BLACK);

  return false;
}

// Performs any necessary cleanup just before the client terminates.
static void cleanup() {
  // restores the terminal to the original mode
  endwin();
}

// Redraws the map on the screen.  This doesn't erase the screen -- it only 
// draws over the map section of the screen.
static void draw_map() {
  // draw the map border
  draw_map_border();

  // draw the map contents
  for (unsigned int row = 0; row < NUM_ROWS; row++) {
    for (unsigned int col = 0; col < NUM_COLS; col++) {
      switch (gs.map_info[row][col]) {
        case MAP_EMPTY:
          attron(COLOR_PAIR(CLR_EMPTY));
          mvaddch(row+MAP_ROW_OFFSET+1, col+MAP_COL_OFFSET+1, ' ');
          attroff(COLOR_PAIR(CLR_EMPTY));
          break;
        case MAP_WALL:
          attron(COLOR_PAIR(CLR_WALL));
          mvaddch(row+MAP_ROW_OFFSET+1, col+MAP_COL_OFFSET+1, '#');
          attroff(COLOR_PAIR(CLR_WALL));
          break;
        case MAP_BRICK:
          attron(COLOR_PAIR(CLR_BRICK));
          mvaddch(row+MAP_ROW_OFFSET+1, col+MAP_COL_OFFSET+1, ACS_CKBOARD);
          attroff(COLOR_PAIR(CLR_BRICK));
          break;
        case MAP_FIRE:
          attron(COLOR_PAIR(CLR_FIRE));
          mvaddch(row+MAP_ROW_OFFSET+1, col+MAP_COL_OFFSET+1, '*');
          attroff(COLOR_PAIR(CLR_FIRE));
          break;
      }
    }
  }
}

// Draws the map border on the screen.
static void draw_map_border() {
  attron(A_BOLD);

  // draw the top and bottom borders
  for (unsigned int col = 1; col < NUM_COLS+1; col++) {
    mvaddch(MAP_ROW_OFFSET, col + MAP_COL_OFFSET, ACS_HLINE);
    mvaddch(MAP_ROW_OFFSET + NUM_ROWS + 1, col + MAP_COL_OFFSET, ACS_HLINE);
  }
  
  // draw the left and right borders
  for (unsigned int row = 1; row < NUM_ROWS+1; row++) {
    mvaddch(row + MAP_ROW_OFFSET, MAP_COL_OFFSET, ACS_VLINE);
    mvaddch(row + MAP_ROW_OFFSET, NUM_COLS + MAP_COL_OFFSET + 1, ACS_VLINE);
  }

  // draw the corners
  mvaddch(MAP_ROW_OFFSET, MAP_COL_OFFSET, ACS_ULCORNER);
  mvaddch(MAP_ROW_OFFSET+NUM_ROWS+1, MAP_COL_OFFSET, ACS_LLCORNER);
  mvaddch(MAP_ROW_OFFSET, MAP_COL_OFFSET+NUM_COLS+1, ACS_URCORNER);
  mvaddch(MAP_ROW_OFFSET+NUM_ROWS+1, MAP_COL_OFFSET+NUM_COLS+1, ACS_LRCORNER);

  attroff(A_BOLD);
}

// Draws the players on the screen.  This should be called after draw_map().
static void draw_players() {
  attron(COLOR_PAIR(CLR_PLAYER) | A_BOLD);
  for (int i = 0; i < NUM_PLAYERS; i++) {
    if (gs.players[i].status == PSTAT_ALIVE) {
      mvaddch(MAP_ROW_OFFSET + 1 + gs.players[i].y_coord, 
              MAP_COL_OFFSET + 1 + gs.players[i].x_coord, 
              (i + '1'));
    }
  }
  attroff(COLOR_PAIR(CLR_PLAYER) | A_BOLD);
}

// Draws the bombs on the screen.  This should be called after draw_players().
static void draw_bombs() {
  int bcnt;
  int pnum;
  int x, y;
  for (int pcnt = 0; pcnt < NUM_PLAYERS; pcnt++) {
    bcnt = gs.players[pcnt].bombs_dropped;
    for (int i = 0; i < bcnt; i++) {
      // if a player is on the same spot, show the player number with a red
      // background
      x = gs.players[pcnt].bombs[i].x_coord;
      y = gs.players[pcnt].bombs[i].y_coord;
      if ((pnum = get_player_at(x, y)) >= 0) {
        attron(COLOR_PAIR(CLR_PLAYBOMB) | A_BOLD);
        mvaddch(MAP_ROW_OFFSET + 1 + y, MAP_COL_OFFSET + 1 + x, (pnum + '1'));
        attroff(COLOR_PAIR(CLR_PLAYBOMB) | A_BOLD);
      } else {
        attron(COLOR_PAIR(CLR_BOMB));
        mvaddch(MAP_ROW_OFFSET + 1 + y, MAP_COL_OFFSET + 1 + x, 'o');
        attroff(COLOR_PAIR(CLR_BOMB));
      }
    }
  }
}

// Draws the game ID, scores, round number, and time left
static void draw_scores() {
  // game ID
  attron(A_BOLD);
  mvprintw(MAP_ROW_OFFSET - 3, MAP_COL_OFFSET, "Game ID: ");
  attroff(A_BOLD);
  printw("%d    ", gs.game_id);

  // player ID
  attron(A_BOLD);
  printw("Player ID: ");
  attroff(A_BOLD);
  printw("%d    ", player_id);

  // server ID
  attron(A_BOLD);
  printw("Server No: ");
  attroff(A_BOLD);
  printw("%d    ", server_id);

  // round number
  attron(A_BOLD);
  mvprintw(MAP_ROW_OFFSET - 2, MAP_COL_OFFSET, "Round: ");
  attroff(A_BOLD);
  printw("%d    ", gs.round_number);

  // time left
  attron(A_BOLD);
  printw("Time: ");
  attroff(A_BOLD);
  printw("%d   ", gs.time_left);

  // scores
  attron(A_BOLD);
  mvprintw(MAP_ROW_OFFSET - 1, MAP_COL_OFFSET, "Scores: ");
  attroff(A_BOLD);
  for (int i = 0; i < NUM_PLAYERS; i++) {
    if (gs.players[i].player_id == player_id) {
      attron(A_BOLD);
      attron(COLOR_PAIR(CLR_PNUM));
      printw("%d=%d  ", i+1, gs.players[i].rounds_won);
      attroff(COLOR_PAIR(CLR_PNUM));
      attroff(A_BOLD);
    } else {
      printw("%d=%d  ", i+1, gs.players[i].rounds_won);
    }
  }
}

// Draw sequence number of player
static void draw_sequence_number(unsigned long num) {
  // sequence number
  attron(A_BOLD);
  mvprintw(MAP_ROW_OFFSET + NUM_ROWS + 2, MAP_COL_OFFSET, "Seq No: ");
  attroff(A_BOLD);
  printw("%lu    ", num);

}

// Updates the timeval struct with the next update time
void set_next_time(struct timeval *next) {
  struct timeval cur;
  gettimeofday(&cur, NULL);

  next->tv_sec   = cur.tv_sec;
  next->tv_usec  = cur.tv_usec;
  next->tv_usec += PERIOD;
  while (next->tv_usec > 1000000) {
    next->tv_sec++;
    next->tv_usec -= 1000000;
  }
}

bool period_expired(const struct timeval *next) {
  struct timeval cur;
  gettimeofday(&cur, NULL);

  if (cur.tv_sec > next->tv_sec) {
    return true;
  } else if ((cur.tv_sec == next->tv_sec) && 
             (cur.tv_usec >= next->tv_usec)) {
    return true;
  }
  return false;
}


/********
 * MAIN *
 ********/

int client_main(int argc, char *argv[], Game_var game, unsigned long playerId, 
                unsigned long gameId, int serverId, 
                CosNaming::NamingContext_var nc) {

  // initialize the client 
  if (initialize()) {
    cout << "Error occurred during initialization" << endl;
    cleanup();
    return -1;
  }

  // Set global variable player_id
  player_id = playerId;

  // Set global variable server_id
  server_id = serverId;

  // initialize the client action structure 
  struct ClientAction action;
  action.seq_num      = 0;
  action.player_id    = playerId; 
  action.arrow        = MOVE_NONE;
  action.request_bomb = false;  
  action.game_id      = gameId;
  assert(action.game_id != 0);

  // timer
  struct timeval time_next;        // next time to send update to the server
  gettimeofday(&time_next, NULL);  // get an update from the server immediately

  // Variables used for retry command
  GameServer_var gameServer;
  CORBA::Object_var gameServerObj;

  // variable used to get the list of gameserver from naming service later
  CosNaming::BindingIterator_var it;  // Iterator reference
  CosNaming::BindingList_var bl;      // Binding list
  const CORBA::ULong CHUNK = MAX_SERVERS;  // max number of supported servers

  // get new list of gameserver from naming service
  nc->list(CHUNK, bl, it);

  // get initial number of server in naming service
  int numServers = bl->length();

  // wait for user input
  int c;
  while ((c = getch()) != '\n') {
    switch (c) {
      case KEY_LEFT:      // move left
        action.arrow = MOVE_LEFT;
        break;
      case KEY_RIGHT:     // move right
        action.arrow = MOVE_RIGHT;
        break;
      case KEY_UP:        // move up
        action.arrow = MOVE_UP;
        break;
      case KEY_DOWN:      // move down
        action.arrow = MOVE_DOWN;
        break;
      case ' ':           // drop bomb
        action.request_bomb = true;
        break;
    }
    

    // send the most recent input to the server
    if (period_expired(&time_next)) {
      // Attempt to update the global GameState object by calling the
      // CORBA remote method
      while(!try_update(game, action)) {
        bool connected = false;

        //if return false, try to reconnect to a different server
        // FIXME: As for now, the client will just keep trying infinitely to any
        // server
        while (!connected) {
          for (int i = 0; i < numServers; i++) {
            try {
              // get new list of gameserver from naming service
              nc->list(CHUNK, bl, it);

              // check number of game server available in naming service,
              // numServers will not decrease, might increase if new server is 
              // being added over time; thus, when bl[i] is being called, it 
              // will not give out of bound error
              numServers = bl->length();

              // Resolving new GameServer from naming service
              gameServerObj = nc->resolve(bl[i].binding_name);
              assert(!CORBA::is_nil(gameServerObj));

              // Adding Game Server reference to gameServer List
              gameServer = GameServer::_narrow(gameServerObj);
              assert(!CORBA::is_nil(gameServer));

              // reconnecting to the new Server
              game = gameServer->retry((CORBA::ULong)playerId,
                  (CORBA::ULong)gameId);
            } catch (const CORBA::Exception & e) {
              // Check the exception to see what it is
              const CORBA::SystemException * se;
              if((se = CORBA::OBJECT_NOT_EXIST::_downcast( &e ))) {
                mvprintw(21, 5, "Game object has been destroyed" );
              } else if((se = CORBA::TRANSIENT::_downcast(&e))) {
                mvprintw(21, 5, "TRANSIENT Exception!" );
              } else if((se = CORBA::COMM_FAILURE::_downcast(&e))) {
                mvprintw(21, 5, "COMM FAILURE!" );
              } else {
                // mvprintw(21, 5, "Unhandled exception: Rethrowing..." );
                // throw;
              }
              continue;
            }
            connected = true;
            // FIXME: As for now, assume gameservers are sorted on ascending 
            // order
            mvprintw(22, 5, "Successfully connected to GameServer %d\n", i+1);
            server_id = i+1;
            break;
          }
        }
      }
      
      if (gs.status == GSTAT_WAITING) {
        // display a message saying waiting for more players
        mvprintw(11, 5, "Game ID: %lu", gs.game_id);
        mvprintw(12, 5, "Waiting for more players (%lu more)", 
                 NUM_PLAYERS - gs.number_players);
        refresh();
        continue;
      } else {
        // draw the over the wait message
        mvprintw(11, 0, "                                      ");
        mvprintw(12, 0, "                                      ");
      }
      
      // reset the action contents
      action.seq_num++;
      action.arrow        = MOVE_NONE;
      action.request_bomb = false;

      // set the next update period
      set_next_time(&time_next);

      // redraw the display with the updated game state
      draw_map();
      draw_players();
      draw_bombs();
      draw_scores();
      draw_sequence_number(action.seq_num);

      mvprintw(23, 5, "press the enter key to exit");

      // update the display
      refresh();
    }
  }

  // perform all cleanup
  cleanup();

  return 0;
}

bool try_update(Game_var game, struct ClientAction &action) {
  try {
    gs = game->update(action);
  } catch (const CORBA::Exception & e) {
    // Check the exception to see what it is
    const CORBA::SystemException * se;
    if( (se = CORBA::OBJECT_NOT_EXIST::_downcast( &e )) ) {
      mvprintw(21, 5, "Game object has been destroyed" );
      return false;
    } else if( (se = CORBA::TRANSIENT::_downcast( &e )) ) {
      mvprintw(21, 5, "TRANSIENT Exception!" );
      return false;
    } else if( (se = CORBA::COMM_FAILURE::_downcast( &e )) ) {
      mvprintw(21, 5, "COMM FAILURE!" );
      // FIXME: need to put something here to handle this exception...
      return false;
    } else {
      // mvprintw(21, 5, "Unhandled exception: Rethrowing..." );
      // throw;
      return false;
    }
  }
  return true;
}

