/* corba_client.cpp
 *
 * The entry point for the client program.  This file performs the CORBA setup 
 * and calls client_main() in another file.
 */


// CORBA includes
#include <OB/CORBA.h>
#include <OB/CosNaming.h>
#include <Game.h>

// standard includes
#include <stdlib.h>
#include <ncurses.h>

// project includes
#include "client.h"

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif


int run(CORBA::ORB_ptr orb, int argc, char* argv[]) {

  // FIXME:: As for now, the user need to specify which server this client needs to connect to
  if(argc != 2) {
    printf("ERROR: Please specify the server ID argument\n");
    return EXIT_FAILURE;
  }

  // Get naming service
  CORBA::Object_var inc;
  try {
    inc = orb->resolve_initial_references("NameService");
  } catch (const CORBA::ORB::InvalidName&) {
    cerr << argv[0] << ": can't resolve `NameService'" <<endl;
    return EXIT_FAILURE;
  }
  if (CORBA::is_nil(inc.in())) {
    cerr << argv[0] << ": `NameService' is a null object reference" <<endl;
    return EXIT_FAILURE;
  }

  CosNaming::NamingContext_var nc = CosNaming::NamingContext::_narrow(inc.in());
  if (CORBA::is_nil(nc.in())) {
    cerr << argv[0] 
         << ": `NameService' is not a NamingContext object reference" << endl;
    return EXIT_FAILURE;
  }

  // Get object by name
  try {
    // Trying to get the whole list of servers from the Naming Service
    CosNaming::BindingIterator_var it;  // Iterator reference
    CosNaming::BindingList_var bl;      // Binding list

    // FIXME:: If there are more than 30 servers, change this to a larger number
    // but this is kinda unlikely given the limited number of game servers we have
    const CORBA::ULong CHUNK = 30;     // Chunk size (get up to 30 servers at a time)
    nc->list(CHUNK, bl, it);
    
    int numServers = bl->length();
    printf("%d GameServers detected in Naming Service\n",numServers);

    GameServer_var gameServer;    
    
    int serverId = atoi(argv[1]);
    // Resolving GameServer from naming service
    // FIXME: Assuming that the server first requested by the client exist, change this to dynamic
    // looking for alive server later
    CORBA::Object_var gameServerObj = nc->resolve(bl[serverId-1].binding_name);
    assert(!CORBA::is_nil(gameServerObj));

    // Adding Game Server reference to gameServer List
    gameServer = GameServer::_narrow(gameServerObj);
    assert(!CORBA::is_nil(gameServer));

    //FIXME: As for now assume there are 2 servers running
    // these values set by join_game
    unsigned int playerId;
    unsigned int gameId;
    // FIXME: handle COMM_Failure exception
    //FIXME: As for now, we define which server this client first connect to
    Game_var game = gameServer->join_game((CORBA::ULong)playerId,
					  (CORBA::ULong)gameId);
    assert(!CORBA::is_nil(game));

    // start the client code
    int rv = client_main(argc, argv, game, playerId, gameId, serverId, nc);
    if (rv < 0) {
      cerr << "Unknown error occurred in client_main" << endl;
      return EXIT_FAILURE;
    }
    
    // player is exiting the game, so notify the server
    game->exit_game(playerId, gameId);

  } catch (const CosNaming::NamingContext::NotFound& ex) {
    cerr << argv[0] << ": Got a `NotFound' exception (";
    switch (ex.why) {
      case CosNaming::NamingContext::missing_node:
        cerr << "missing node";
        break;
      case CosNaming::NamingContext::not_context:
        cerr << "not context";
        break;
      case CosNaming::NamingContext::not_object:
        cerr << "not object";
        break;
    }
    cerr << ")" << endl;
    return EXIT_FAILURE;
  } catch (const CosNaming::NamingContext::CannotProceed&) {
    cerr << argv[0] << ": Got a `CannotProceed' exception" << endl;
    return EXIT_FAILURE;
  } catch (const CosNaming::NamingContext::InvalidName&) {
    cerr << argv[0] << ": Got an `InvalidName' exception" << endl;
    return EXIT_FAILURE;
  } catch (const CosNaming::NamingContext::AlreadyBound&) {
    cerr << argv[0] << ": Got an `AlreadyBound' exception" << endl;
    return EXIT_FAILURE;
  } catch (const CosNaming::NamingContext::NotEmpty&) {
    cerr << argv[0] << ": Got a `NotEmpty' exception" << endl;
    return EXIT_FAILURE;
  }
  return EXIT_SUCCESS;
}


int main(int argc, char* argv[], char*[]) {
  int status = EXIT_SUCCESS;
  CORBA::ORB_var orb;

  // This try/catch block mainly catches exceptions thrown by
  // Game::exit_game().  Also catches other generic failures
  try {
    orb = CORBA::ORB_init(argc, argv, "Game-Client");
    status = run(orb, argc, argv);
  } catch (const CORBA::Exception & e) {
    // Check the exception to see what it is
    const CORBA::SystemException * se;
    if( (se = CORBA::OBJECT_NOT_EXIST::_downcast( &e )) ) {
      printf( "Game object has been destroyed\n" );
    } else if( (se = CORBA::TRANSIENT::_downcast( &e )) ) {
      printf( "TRANSIENT Exception!\n" );
    } else if( (se = CORBA::COMM_FAILURE::_downcast( &e )) ) {
      printf( "COMM FAILURE!\n" );
    } else {
      // mvprintw(22, 5, "Rethrowing exception..." );
      // throw;
      cerr << e << endl;
      status = EXIT_FAILURE;
    }
  }

  if (!CORBA::is_nil(orb)) {
    try {
      orb->destroy();
    } catch (const CORBA::Exception& ex) {
      cerr << ex << endl;
      status = EXIT_FAILURE;
    }
  }

  return status;
}
