/************
 * INCLUDES *
 ************/

// CORBA includes
#include <OB/CORBA.h>

// project includes
#include <Game_impl.h>
#include "sql_lib.h"
#include "game_constants.h"

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

// ----------------------------------------------------------------------
// Game_impl constructor
// ----------------------------------------------------------------------

Game_impl::Game_impl(CORBA::ORB_ptr orb, PortableServer::POA_ptr poa,
                     CORBA::ULong gid)
    : orb_(CORBA::ORB::_duplicate(orb)), 
      poa_(PortableServer::POA::_duplicate(poa))
{
  // initialize the game state
  gs.game_id = game_id = gid;

  // connect the database
  db_connect();
  
  // FIXME: Why do we need to create a new master table everytime
  // create the database table
  db_create_master_table();

  db_get_gamestate(game_id, &gs);

  // initialize the game state data for the first round
  //FIXME: game should not be initialized totally if the game
  // already exist on the database
  // this is so because this constructor might be called by a client
  // rejoining a new server, thus the game state should not be reinitialized
  if(gs.seq_num == 0)
    this->init_round();  // FIXME: do this here?

  // set these values in the database
  db_set_gamestate(&gs);

} 
// ----------------------------------------------------------------------
// Game_impl Destructor
// ----------------------------------------------------------------------

Game_impl::~Game_impl() {
  db_disconnect();
}


// ----------------------------------------------------------------------
// Game_impl private member implementation
// ----------------------------------------------------------------------

void Game_impl::init_round() {
  static int round_num = 1;
  char map_str[NUM_ROWS * NUM_COLS + 1];
  exploders.clear();
  bomb_loc.clear();
  
  PlayerInfo *current;
  // Reset everything!
  for( int i = 0; i < NUM_PLAYERS; i++ ) {
    current = &gs.players[i];
    if( current->player_id != INVALID_ID ) {
      if( current->status & PSTAT_DEAD ) {
	// Remove PSTAT_DEAD flag.
	current->status &= ~PSTAT_DEAD; 
	for( int j = 0; j < current->bombs_dropped; j++ ) {
	  current->bombs[j].status = EXPLODED;
	}
	current->bombs_dropped = 0;
      }
    }
  }

  // set the round number and remaining time
  gs.round_number = round_num++;
  gs.time_left    = ROUND_TIME;

  // initialize the map values
  strcpy(map_str, "002222222222200012121212121210222222222222222212121212121212222222222222222212121212121212222222222222222212121212121212222222222222222012121212121210002222222222200");
  map_str_to_array(map_str, &gs);

  // reset player locations
  gs.players[0].x_coord = 0;
  gs.players[0].y_coord = 0;
  gs.players[1].x_coord = NUM_COLS - 1;
  gs.players[1].y_coord = 0;
  gs.players[2].x_coord = 0;
  gs.players[2].y_coord = NUM_ROWS - 1;
  gs.players[3].x_coord = NUM_COLS - 1;
  gs.players[3].y_coord = NUM_ROWS - 1;

}

struct PlayerInfo* Game_impl::get_player(unsigned int pid) {
  for (int i = 0; i < NUM_PLAYERS; i++) {
    if (gs.players[i].player_id == pid) {
      return &gs.players[i];
    }
  }
  return NULL;
}

int Game_impl::check_for_winner() {
  // FIXME: use NUM_WINS constant to check for winner
  int live_players = 0;
  struct PlayerInfo *winner = NULL;
  for(int i = 0; i < NUM_PLAYERS; i++) {
    if( !(gs.players[i].status & PSTAT_DEAD) ) {
      winner = &gs.players[i];
      live_players++;
    }
  }
  if( live_players <= 1 ){
    // Round has ended
    if( live_players == 1 ) {
      winner->rounds_won++;
    }
    
    //    round_num++;
    //gs.round_number++;

    // Reset the round
    this->init_round();
  }

  return 0;
}
	  
#define to_hash(x,y) (10000*(x) + (y))
#define x_from_hash(x) ((x)/10000)
#define y_from_hash(x) ((x)%10000)

void Game_impl::resolve_fire() {

  bomb_loc.clear();
  exploders.clear();
  exploders.reserve( NUM_PLAYERS * MAX_BOMBS );

  struct PlayerInfo *pi = NULL;
  int player_number = 0;
  int x_coord;
  int y_coord;

  for( int i = 0; i < NUM_PLAYERS; i++ ) {
    pi = &(gs.players[i]);

    if( pi->player_id == INVALID_ID )
      break;

    // player_number CAN'T be used to offset into gs.players[NUM_PLAYERS]!!!!
    player_number = pi - &(gs.players[0]) + 1;
    int exploded_bombs = 0;
    for(int j = 0; j < pi->bombs_dropped; j++ ) {
      // Sanity check
      if( pi->bombs[j].bomb_id == INVALID_ID || j >= MAX_BOMBS )
        break;
      x_coord = pi->bombs[j].x_coord;
      y_coord = pi->bombs[j].y_coord;
      bomb_loc[ to_hash(x_coord, y_coord) ] = &(pi->bombs[j]);

      // Check bomb to see if its ready to explode
      if( pi->bombs[j].det_seq_num[player_number-1] <= pi->seq_num ) {
        pi->bombs[j].status = EXPLODED;
        exploded_bombs += 1;
        assert( gs.players[i].bombs_dropped >= 0);
        // Add bomb to list of bombs to explode
        exploders.push_back( &(pi->bombs[j]) );
      }
    }
    pi->bombs_dropped -= exploded_bombs;
  }

  if ( exploders.empty() ) {
    return;
  }
  
  for( exploders_iter = exploders.begin(); 
       exploders_iter != exploders.end(); 
       exploders_iter++ ) {
    // Resolve fire radius for NO movement.
    resolve_fire_radius( FIRE_RADIUS, MOVE_UP );
    resolve_fire_radius( FIRE_RADIUS, MOVE_DOWN );
    resolve_fire_radius( FIRE_RADIUS, MOVE_LEFT );
    resolve_fire_radius( FIRE_RADIUS, MOVE_RIGHT );
  }
}

void Game_impl::resolve_fire_radius( int radius, int direction )
{
  int x_movement;
  int y_movement;

  switch( direction ) {
    case MOVE_UP:
      x_movement = 0;
      y_movement = -1;
      break;
    case MOVE_DOWN:
      x_movement = 0;
      y_movement = 1;
      break;
    case MOVE_RIGHT:
      x_movement = 1;
      y_movement = 0;
      break;
    case MOVE_LEFT:
      x_movement = -1;
      y_movement = 0;
      break;
    default:
      // Oops.
      return;
  };

  int x_coord = (*exploders_iter)->x_coord;
  int y_coord = (*exploders_iter)->y_coord;
  for( int i = 0; i < radius; i++ ) {
    x_coord += x_movement;
    y_coord += y_movement;
    
    // Corner cases: can't bomb off the map.
    if( x_coord < 0 || x_coord >= NUM_COLS )
      break;
    if( y_coord < 0 || y_coord >= NUM_ROWS )
      break;

    long int *map_point = &( gs.map_info[y_coord][x_coord] );

    if( *map_point == MAP_EMPTY ) {
      // See if a hapless player is standing in the line of fire
      for( int j = 0; j < NUM_PLAYERS; j++ ) {
        if( gs.players[j].x_coord == x_coord &&
            gs.players[j].y_coord == y_coord ) {
          gs.players[j].status |= PSTAT_DEAD;
          cout << "Killing player " << j+1 << endl;
        }
      }
      // See if this bomb set off other bombs
      bomb_loc_iter = bomb_loc.find( to_hash( x_coord, y_coord ) );
      if( bomb_loc_iter != bomb_loc.end() ) {
        // Push new bomb onto stack of exploders
        assert( x_coord == bomb_loc_iter->second->x_coord );
        assert( y_coord == bomb_loc_iter->second->y_coord );
        if( bomb_loc_iter->second->status == UNEXPLODED ) {
          bomb_loc_iter->second->status = EXPLODED;
          //FIXME: Gotta decremeted a player's bomb count
          exploders.push_back( bomb_loc_iter->second );
        }
        break;
      } else {
        continue;
      }
    } else if( *map_point == MAP_WALL ) {
      break;
    } else if( *map_point == MAP_BRICK ) {
      // Brick is destroyed
      *map_point = MAP_EMPTY;
      break;
    }
  }
}
	 


// ----------------------------------------------------------------------
// Game_impl public member implementation
// ----------------------------------------------------------------------

struct GameState
Game_impl::update(const struct ClientAction &c_action)
    throw(CORBA::SystemException)
{
  static unsigned long seq = 0;

  // Get the current game state from database
  db_get_gamestate(game_id, &gs);
  
  // find the player that corresponds to the client that called update()
  struct PlayerInfo *pi = get_player(c_action.player_id);
  if (pi == NULL) {
    cerr << "ERROR: INVALID PLAYER ID";
    return gs;
  }
  
  unsigned int x   = pi->x_coord;
  unsigned int y   = pi->y_coord;
  unsigned int pid = pi->player_id;
  unsigned int player_number = pi - &gs.players[0] + 1;
  // player_number CAN'T be used to offset into gs.players[NUM_PLAYERS]!!!!

  // See if the sequence number is valid: bail if it's not
  // FIXME: There might be a difference in the sequence number if the server 
  // dies after incrementing the seq number but the client did not
  /*
  if(!((c_action.seq_num == pi->seq_num) || 
       (c_action.seq_num+1 == pi->seq_num) || 
       (c_action.seq_num-1 == pi->seq_num))) {
  */
  if( c_action.seq_num != pi->seq_num) {    
    pi->seq_num += 1;
    return gs;
  }

  // fill in the GameState structure to send back to the client
  gs.seq_num = seq++;

  // Dead players can't really do much.
  if (pi->status & PSTAT_DEAD) {
    pi->seq_num += 1;
    db_set_gamestate(&gs);
    return gs;
  }

  // plant a bomb if necessary
  if (c_action.request_bomb) {
    // FIXME: Can currently overwrite other bombs.
    if (pi->bombs_dropped < MAX_BOMBS) {
      pi->bombs[pi->bombs_dropped].player_id = pid;
      pi->bombs[pi->bombs_dropped].status = NEW;
      pi->bombs[pi->bombs_dropped].x_coord = x;
      pi->bombs[pi->bombs_dropped].y_coord = y;
      pi->bombs[pi->bombs_dropped].det_seq_num[0] = gs.players[0].seq_num + 
                                                    BOMB_TIMER;
      pi->bombs[pi->bombs_dropped].det_seq_num[1] = gs.players[1].seq_num + 
                                                    BOMB_TIMER;
      pi->bombs[pi->bombs_dropped].det_seq_num[2] = gs.players[2].seq_num + 
                                                    BOMB_TIMER;
      pi->bombs[pi->bombs_dropped].det_seq_num[3] = gs.players[3].seq_num + 
                                                    BOMB_TIMER;
      pi->bombs_dropped++;
    }
  }

  // handle the player movement
  switch (c_action.arrow) {
    case MOVE_LEFT:
      if ((x > 0) && (gs.map_info[y][x-1] == MAP_EMPTY)) {
        pi->x_coord = pi->x_coord - 1;
      }
      break;
    case MOVE_RIGHT:
      if ((x < NUM_COLS-1) && (gs.map_info[y][x+1] == MAP_EMPTY)) {
        pi->x_coord = pi->x_coord + 1;
      }
      break;
    case MOVE_UP:
      if ((y > 0) && (gs.map_info[y-1][x] == MAP_EMPTY)) {
        pi->y_coord = pi->y_coord - 1;
      }
      break;
    case MOVE_DOWN:
      if ((y < NUM_ROWS-1) && (gs.map_info[y+1][x] == MAP_EMPTY)) {
        pi->y_coord = pi->y_coord + 1;
      }
      break;
  }

  // blows up bombs, kills players, and destroys bricks
  this->resolve_fire();

  // Increment client's seq_num for use the next time.
  pi->seq_num += 1;

  // reset round and update score if the round is done
  this->check_for_winner();

  // write the gs values back to the database
  db_set_gamestate(&gs);

  return gs;
}

unsigned long Game_impl::exit_game(CORBA::ULong player_id, CORBA::ULong game_id)
  throw(CORBA::SystemException)
{
  // removes a player from the game (marks him as no longer in the game)
  int flag = db_exit_game(game_id, player_id);

  // get playerInfo
  struct PlayerInfo *pi = get_player(player_id);
  if (pi == NULL) {
    cerr << "ERROR: INVALID PLAYER ID";
    return 0;
  }

  // FIXME: remove this
  cout << "Player " << player_id << " from Game " 
       << game_id << " left the game" << endl;

  // FIXME: need to test this
  if (flag == -2) {
    // Get the default POA to associate this object to
    PortableServer::POA_var poa = _default_POA();

    char gameIdStr[10];
    PortableServer::ObjectId_var gameId = 
      PortableServer::string_to_ObjectId(gameIdStr);
    poa->deactivate_object(gameId); 
    cout << "Game " << game_id << " deactivated" << endl;
  }

  return player_id;
}

PortableServer::POA_ptr
Game_impl::_default_POA()
{
    return PortableServer::POA::_duplicate(poa_);
}

