#include <OB/CORBA.h>
#include <OB/BootManager.h>

#include <Game_impl.h>
#include <GameServer_impl.h>

#include <stdlib.h>
#include <errno.h>

#include "sql_lib.h"

#ifdef HAVE_FSTREAM
#   include <fstream>
#else
#   include <fstream.h>
#endif

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif

#include <OB/CosNaming.h>

int run(CORBA::ORB_ptr orb, int argc, char* argv[]) {
  
  // FIXME: as for now, the server will take in it's server ID
  if(argc != 2) {
    printf("ERROR: Please include server Number\n");
    return EXIT_FAILURE;
  }
  
  printf("Server %d:\n",atoi(argv[1]));

  // Resolve Root POA
  CORBA::Object_var poaObj = orb->resolve_initial_references("RootPOA");
  PortableServer::POA_var rootPOA = PortableServer::POA::_narrow(poaObj);
  assert(!CORBA::is_nil(rootPOA));

  // Create NO_IMPLICIT_ACTIVATION policy object
  PortableServer::ImplicitActivationPolicy_var implicitActive = 
    rootPOA -> create_implicit_activation_policy(PortableServer::NO_IMPLICIT_ACTIVATION);

  // Create USER_ID policy object
  PortableServer::IdAssignmentPolicy_var idAssignment = 
    rootPOA -> create_id_assignment_policy(PortableServer::USER_ID);

  // Create RETAIN policy object
  PortableServer::ServantRetentionPolicy_var servantRetention =
    rootPOA -> create_servant_retention_policy(PortableServer::RETAIN);

  // Make the POA single threaded
  PortableServer::ThreadPolicy_var threadPolicy =
    rootPOA -> create_thread_policy(PortableServer::SINGLE_THREAD_MODEL);

  // Create PolicyList for child POA
  CORBA::PolicyList gamePoaPolicies;
  gamePoaPolicies.length(4); 
  gamePoaPolicies[0] = PortableServer::ImplicitActivationPolicy::_duplicate(implicitActive);
  gamePoaPolicies[1] = PortableServer::IdAssignmentPolicy::_duplicate(idAssignment);
  gamePoaPolicies[2] = PortableServer::ServantRetentionPolicy::_duplicate(servantRetention);
  gamePoaPolicies[3] = PortableServer::ThreadPolicy::_duplicate(threadPolicy);

  // Set up a nil POAManager reference to pass to the GamePOA create_POA call
  PortableServer::POAManager_var nil_mgr = 
    PortableServer::POAManager::_nil();

  // Create GamePOA
  PortableServer::POA_var gamePOA =
    rootPOA -> create_POA("GamePOA", nil_mgr, gamePoaPolicies);

  // Destroy all policies
  implicitActive->destroy();
  idAssignment->destroy();
  servantRetention->destroy();
  threadPolicy->destroy();

  // Get a reference to the POA manager
  PortableServer::POAManager_var manager = gamePOA->the_POAManager();

  // Create implementation GameServer object
  GameServer_impl* gameServerImpl = new GameServer_impl(orb, gamePOA);
  PortableServer::ServantBase_var gameServer_servant = gameServerImpl;
  //GameServer_var gameServer = gameServerImpl->_this();

  // Activate gameServer object to gamePOA
  PortableServer::ObjectId_var gameServerId = 
    PortableServer::string_to_ObjectId("MainGameServer"); 
  gamePOA -> activate_object_with_id(gameServerId, gameServerImpl);
  
  // Obtain gameServer object reference
  CORBA::Object_var gameServerObj = gamePOA->id_to_reference(gameServerId);
  GameServer_var gameServer = GameServer::_narrow(gameServerObj);

  // Get naming service
  CORBA::Object_var inc;

  try {
    inc = orb->resolve_initial_references("NameService");
  } 
  catch (const CORBA::ORB::InvalidName&) {
    cerr << argv[0] << ":can't resolve `NameService'" << endl;
    return EXIT_FAILURE;
  }
  
  if (CORBA::is_nil(inc.in())) {
    cerr << argv[0] << ": `NameService' is a nil object reference" << endl;
    return EXIT_FAILURE;
  }
  
  CosNaming::NamingContext_var nc = CosNaming::NamingContext::_narrow(inc.in());

  if (CORBA::is_nil(nc.in())) {
    cerr << argv[0] << ": `NameService is not a NamingContext object reference"          << endl;
    return EXIT_FAILURE;
  }

  // Check if this server is the first server to connect to the Naming Service
  // if so, this server will be the server to initialize the database first later
  CosNaming::BindingIterator_var it;  // Iterator reference
  CosNaming::BindingList_var bl;      // Binding list

  // FIXME:: If there are more than 30 servers, change this to a larger number
  // but this is kinda unlikely given the limited number of game servers we have
  const CORBA::ULong CHUNK = 30;     // Chunk size (get up to 30 servers at a time)
  nc->list(CHUNK, bl, it);
  
  // connect to database
  db_connect();

  if(bl->length() == 0 && atoi(argv[1]) == 1) {
    printf("First Server initialized: Creating database\n");
    db_drop_all_tables();
    db_create_master_table();
  }

  try {
    // Bind name with the Naming Service
    CosNaming::Name gameServerName;
    gameServerName.length(1);
    
    char buffer[15];
    sprintf(buffer,"GameServer%s",argv[1]);
    gameServerName[0].id = CORBA::string_dup(buffer);

    printf("Binding %s to Naming Service\n",buffer);
    try {
      nc->rebind(gameServerName,gameServer.in());
    }
    catch (const CosNaming::NamingContext::NotFound &) {
      printf("Naming Service throws NotFound Exception");
      return EXIT_FAILURE;
    }
    catch (const CosNaming::NamingContext::CannotProceed &) {
      printf("Naming Service throws CannotProceed Exception");
      return EXIT_FAILURE;
    }
    catch (const CosNaming::NamingContext::InvalidName &) {
      printf("Naming Service throws InvalidName Exception");
      return EXIT_FAILURE;
    }

    manager->activate();
    orb->run();

    // unbind object from naming service
    nc->unbind(gameServerName);
  } catch (const CosNaming::NamingContext::NotFound& ex) {
    cerr << argv[0] << ": Got a `NotFound' exception (";
    switch (ex.why) {
      case CosNaming::NamingContext::missing_node:
        cerr << "missing node";
        break;
      case CosNaming::NamingContext::not_context:
        cerr << "not context";
        break;
      case CosNaming::NamingContext::not_object:
        cerr << "not object";
        break;
    }
    cerr << ")" << endl;
    return EXIT_FAILURE;
  } catch (const CosNaming::NamingContext::CannotProceed&) {
    cerr << argv[0] << ": Got a `CannotProceed' exception" << endl;
    return EXIT_FAILURE;
  } catch (const CosNaming::NamingContext::InvalidName&) {
    cerr << argv[0] << ": Got an `InvalidName' exception" << endl;
    return EXIT_FAILURE;
  } catch (const CosNaming::NamingContext::AlreadyBound&) {
    cerr << argv[0] << ": Got an `AlreadyBound' exception" << endl;
    return EXIT_FAILURE;
  } catch (const CosNaming::NamingContext::NotEmpty&) {
    cerr << argv[0] << ": Got a `NotEmpty' exception" << endl;
    return EXIT_FAILURE;
  }
  return EXIT_SUCCESS;
}

int main(int argc, char* argv[], char*[]) {
  int status = EXIT_SUCCESS;
  CORBA::ORB_var orb;

  try {
    orb = CORBA::ORB_init(argc, argv);
    status = run(orb, argc, argv);
  } catch(const CORBA::Exception& ex) {
    cerr << ex << endl;
    status = EXIT_FAILURE;
  }

  if (!CORBA::is_nil(orb)) {
    try {
      orb -> destroy();
    } catch(const CORBA::Exception& ex) {
      cerr << ex << endl;
      status = EXIT_FAILURE;
    }
  }

  return status;
}

